
import redis
from assemblyline.remote.datatypes import get_client, retry_call

begin_script = """
local t = redis.call('time')
local key = tonumber(t[1] .. string.format("%06d", t[2]))

local name = ARGV[1]
local max = tonumber(ARGV[2])
local timeout = tonumber(ARGV[3] .. "000000")

redis.call('zremrangebyscore', name, 0, key - timeout)
if redis.call('zcard', name) < max then
    redis.call('zadd', name, key, key)
    return true
else
    return false
end
"""


class UserQuotaTracker(object):
    def __init__(self, prefix, timeout=120, redis=None, host=None, port=None, private=False):
        self.c = redis or get_client(host, port, private)
        self.bs = self.c.register_script(begin_script)
        self.prefix = prefix
        self.timeout = timeout

    def _queue_name(self, user):
        return f"{self.prefix}-{user}"

    def begin(self, user, max_quota):
        try:
            return retry_call(self.bs, args=[self._queue_name(user), max_quota, self.timeout]) == 1
        except redis.exceptions.ResponseError as er:
            # TODO: This is a failsafe for upgrade purposes could be removed in a future version
            if "WRONGTYPE" in str(er):
                retry_call(self.c.delete, self._queue_name(user))
                return retry_call(self.bs, args=[self._queue_name(user), max_quota, self.timeout]) == 1
            else:
                raise

    def end(self, user):
        """When only one item is requested, blocking is is possible."""
        try:
            retry_call(self.c.zpopmin, self._queue_name(user))
        except redis.exceptions.ResponseError as er:
            # TODO: This is a failsafe for upgrade purposes could be removed in a future version
            if "WRONGTYPE" in str(er):
                retry_call(self.c.delete, self._queue_name(user))
                retry_call(self.c.zpopmin, self._queue_name(user))
            else:
                raise



import logging
import os
import shutil
import re
from typing import AnyStr, Iterable, Optional

from assemblyline.common.exceptions import ChainAll
from assemblyline.common.path import strip_path_inclusion
from assemblyline.common.uid import get_random_id
from assemblyline.filestore.transport.base import Transport, TransportException, normalize_srl_path

NORMALIZED = re.compile('[a-z0-9]/[a-z0-9]/[a-z0-9]/[a-z0-9]/[a-z0-9]{64}')


@ChainAll(TransportException)
class TransportLocal(Transport):
    """
    Local file system Transport class.
    """

    def __init__(self, base=None, normalize=None):
        self.log = logging.getLogger('assemblyline.transport.local')
        self.base = base
        self.host = "localhost"

        def local_normalize(path):
            # If they've provided an absolute path. Leave it a is.
            if path.startswith('/'):
                s = path
            # Relative paths
            elif '/' in path or len(path) != 64:
                s = _join(self.base, path)
            else:
                s = _join(self.base, normalize_srl_path(path))
            self.log.debug('local normalized: %s -> %s', path, s)
            return s

        if not normalize:
            normalize = local_normalize

        super(TransportLocal, self).__init__(normalize=normalize)

    def delete(self, path):
        normal_path = self.normalize(path)
        try:
            os.unlink(normal_path)
        except FileNotFoundError:
            pass
        except OSError as error:
            raise ValueError(f"Error erasing {path} as {normal_path}: {error}") from error

    def exists(self, path):
        path = self.normalize(path)
        return os.path.exists(path)

    def makedirs(self, path):
        path = self.normalize(path)
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno == 17:
                pass
            else:
                raise e

    # File based functions
    def download(self, src_path, dst_path):
        if src_path == dst_path:
            return

        src_path = self.normalize(src_path)
        dir_path = os.path.dirname(dst_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        shutil.copy(src_path, dst_path)

    def upload(self, src_path, dst_path):
        dst_path = self.normalize(dst_path)
        if src_path == dst_path:
            return

        dirname = os.path.dirname(dst_path)
        filename = os.path.basename(dst_path)
        tempname = get_random_id()
        temppath = _join(dirname, tempname)
        finalpath = _join(dirname, filename)
        assert (finalpath == dst_path)
        self.makedirs(dirname)
        shutil.copy(src_path, temppath)
        shutil.move(temppath, finalpath)
        assert (self.exists(dst_path))

    # Buffer based functions
    def get(self, path: str) -> bytes:
        path = self.normalize(path)
        fh = None
        try:
            fh = open(path, "rb")
            return fh.read()
        finally:
            if fh:
                fh.close()

    def put(self, path: str, content: AnyStr):
        path = self.normalize(path)

        dirname = os.path.dirname(path)
        filename = os.path.basename(path)

        tempname = get_random_id()
        temppath = _join(dirname, tempname)

        finalpath = _join(dirname, filename)
        assert(finalpath == path)

        self.makedirs(dirname)
        fh = None
        try:
            fh = open(temppath, "wb")
            if isinstance(content, str):
                return fh.write(content.encode())
            else:
                return fh.write(content)
        finally:
            if fh:
                fh.close()

            try:
                shutil.move(temppath, finalpath)
            except shutil.Error:
                pass
            assert(self.exists(path))

    def __str__(self):
        return 'file://{}'.format(self.base)

    def _denormalize(self, name):
        if NORMALIZED.fullmatch(name):
            return name.split('/')[-1]
        return name

    def list(self, prefix: Optional[str] = None) -> Iterable[str]:
        for name in self._list(self.base, prefix, prefix):
            yield self._denormalize(name)

    def _list(self, path: str, dir_prefix: Optional[str], file_prefix: Optional[str]) -> Iterable[str]:
        for listed in os.listdir(path):
            listed_path = os.path.join(path, listed)
            if os.path.isdir(listed_path):
                if dir_prefix:
                    if listed.startswith(dir_prefix) or dir_prefix.startswith(listed):
                        for subfile in self._list(listed_path, dir_prefix[len(listed):], file_prefix):
                            yield os.path.join(listed, subfile)
                else:
                    for subfile in self._list(listed_path, None, file_prefix):
                        yield os.path.join(listed, subfile)
            if os.path.isfile(listed_path):
                if not file_prefix or listed.startswith(file_prefix) or not dir_prefix or listed.startswith(dir_prefix):
                    yield listed


###############################
# Helper functions.
###############################


def _join(base, path):
    path = path.replace("\\", "/").replace("//", "/")
    if base is None:
        return path
    return os.path.join(base, strip_path_inclusion(path, base)).replace("\\", "/")



# This file contains the loaders for the different components of the system
from __future__ import annotations

import elasticapm
import importlib
import os
import time
import yaml

from string import Template
from typing import TYPE_CHECKING

from assemblyline.common.constants import service_queue_name
from assemblyline.common.dict_utils import recursive_update
from assemblyline.common.importing import load_module_by_path

if TYPE_CHECKING:
    from assemblyline.odm.models.config import Config

config_cache = {}


def get_apm_client(service_name):
    config = get_config()
    apm_config = dict(server_url=config.core.metrics.apm_server.server_url, service_name=service_name)
    if config.core.metrics.apm_server.server_url.startswith('https'):
        # Due to https://github.com/elastic/apm-agent-python/blob/208e241fbf28b4bb09ba5aca3dbd5f7b95602229/elasticapm/transport/http.py#L250
        # We'll disable certificate verification from the APM server
        apm_config['verify_server_cert'] = False

    return elasticapm.Client(**apm_config)


def get_classification(yml_config=None):
    from assemblyline.common.classification import Classification, InvalidDefinition

    if yml_config is None:
        yml_config = "/etc/assemblyline/classification.yml"

    classification_definition = {}
    default_file = os.path.join(os.path.dirname(__file__), "classification.yml")
    if os.path.exists(default_file):
        with open(default_file) as default_fh:
            default_yml_data = yaml.safe_load(default_fh.read())
            if default_yml_data:
                classification_definition.update(default_yml_data)

    # Load modifiers from the yaml config
    if os.path.exists(yml_config):
        with open(yml_config) as yml_fh:
            yml_data = yaml.safe_load(yml_fh.read())
            if yml_data:
                classification_definition = recursive_update(classification_definition, yml_data)

    if not classification_definition:
        raise InvalidDefinition('Could not find any classification definition to load.')

    return Classification(classification_definition)


def env_substitute(buffer):
    """Replace environment variables in the buffer with their value.

    Use the built in template expansion tool that expands environment variable style strings ${}
    We set the idpattern to none so that $abc doesn't get replaced but ${abc} does.

    Case insensitive.
    Variables that are found in the buffer, but are not defined as environment variables are ignored.
    """
    return Template(buffer).safe_substitute(os.environ, idpattern=None, bracedidpattern='(?a:[_a-z][_a-z0-9]*)')


def _get_config(yml_config=None):
    from assemblyline.odm.models.config import Config

    if yml_config is None:
        yml_config = "/etc/assemblyline/config.yml"

    # Initialize a default config
    config = Config().as_primitives()

    # Load modifiers from the yaml config
    if os.path.exists(yml_config):
        with open(yml_config) as yml_fh:
            yml_data = yaml.safe_load(env_substitute(yml_fh.read()))
            if yml_data:
                config = recursive_update(config, yml_data)

    if 'AL_LOG_LEVEL' in os.environ:
        config['logging']['log_level'] = os.environ['AL_LOG_LEVEL']

    return Config(config)


def get_archivestore(config=None, connection_attempts=None):
    from assemblyline.filestore import FileStore
    if config is None:
        config = get_config()
    if config.datastore.archive.enabled:
        archive_storage_url = config.filestore.archive or config.filestore.storage
        return FileStore(*archive_storage_url, connection_attempts=connection_attempts)
    else:
        raise ValueError("Trying to access the archive filestore but archive is disabled.")


def get_config(yml_config=None) -> Config:
    if yml_config not in config_cache:
        config_cache[yml_config] = _get_config(yml_config=yml_config)
    return config_cache[yml_config]


def get_constants(config=None):
    if config is None:
        config = get_config()
    return importlib.import_module(config.system.constants)


def get_datastore(config=None, archive_access=False):
    from assemblyline.datastore.helper import AssemblylineDatastore
    from assemblyline.datastore.store import ESStore

    if not config:
        config = get_config()

    return AssemblylineDatastore(ESStore(config.datastore.hosts, archive_access=archive_access))


def get_cachestore(component, config=None, datastore=None):
    from assemblyline.cachestore import CacheStore
    return CacheStore(component, config=config, datastore=datastore)


def get_filestore(config=None, connection_attempts=None):
    from assemblyline.filestore import FileStore
    if config is None:
        config = get_config()
    return FileStore(*config.filestore.storage, connection_attempts=connection_attempts)


def get_identify(use_cache=True, config=None, datastore=None, log=None):
    from assemblyline.common.identify import Identify
    return Identify(use_cache=use_cache, config=config, datastore=datastore, log=log)


def get_process_alert_message():
    config = get_config()
    return load_module_by_path(config.core.alerter.process_alert_message)


def get_metrics_sink(redis=None):
    from assemblyline.remote.datatypes.queues.comms import CommsQueue
    return CommsQueue('assemblyline_metrics', host=redis)


def get_service_queue(service: str, redis=None):
    from assemblyline.remote.datatypes.queues.priority import PriorityQueue
    return PriorityQueue(service_queue_name(service), redis)


def get_tag_safelist_data(yml_config=None):

    if yml_config is None:
        yml_config = "/etc/assemblyline/tag_safelist.yml"

    tag_safelist_data = {}
    default_file = os.path.join(os.path.dirname(__file__), "tag_safelist.yml")
    if os.path.exists(default_file):
        with open(default_file) as default_fh:
            default_yml_data = yaml.safe_load(default_fh.read())
            if default_yml_data:
                tag_safelist_data.update(default_yml_data)

    # Load modifiers from the yaml config
    if os.path.exists(yml_config):
        with open(yml_config) as yml_fh:
            yml_data = yaml.safe_load(yml_fh.read())
            if yml_data:
                tag_safelist_data = recursive_update(tag_safelist_data, yml_data)

    return tag_safelist_data


def get_tag_safelister(log=None, yml_config=None, config=None, datastore=None):
    from assemblyline.common.tagging import TagSafelister, InvalidSafelist

    with get_cachestore('system', config=config, datastore=datastore) as cache:
        tag_safelist_yml = cache.get('tag_safelist_yml')
        if tag_safelist_yml:
            tag_safelist_data = yaml.safe_load(tag_safelist_yml)
        else:
            tag_safelist_data = get_tag_safelist_data(yml_config=yml_config)

    if not tag_safelist_data:
        raise InvalidSafelist('Could not find any tag_safelist file to load.')

    return TagSafelister(tag_safelist_data, log=log)


class CachedObject:
    """An object proxy that automatically refreshes its target periodically."""

    def __init__(self, factory, refresh=None, args=None, kwargs=None):
        """
        Args:
            factory: Factory that takes the arguments given in `args` and `kwargs` and produces the proxyed object.
            refresh: Refresh interval in seconds.
        """
        self.__factory = factory
        self.__refresh = float(refresh or 60)
        self.__cached = None
        self.__update_time = 0
        self.__args = args or []
        self.__kwargs = kwargs or {}

    def __reload(self):
        if time.time() - self.__update_time > self.__refresh:
            with elasticapm.capture_span(name=f"CachedObject.reload({self.__factory.__name__})",
                                         span_type="cached_object"):
                self.__cached = self.__factory(*self.__args, **self.__kwargs)
                self.__update_time = time.time()

    def __getattr__(self, key):
        """Forward all attribute requests to the underlying object.

        Refresh the object every `_update_time` seconds.
        """
        self.__reload()
        return getattr(self.__cached, key)

    def __getitem__(self, item):
        self.__reload()
        return self.__cached[item]



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


@odm.model(index=False, store=False, description="Abstract Model of Favorite")
class Favorite(odm.Model):
    created_by = odm.Keyword(description="Who created the favorite")
    classification = odm.Classification(is_user_classification=True, copyto="__text__",
                                        default=Classification.UNRESTRICTED,
                                        description="Classification of the favorite")
    name = odm.Keyword(description="Name of the favorite")
    query = odm.Keyword(description="Query for the favorite")


@odm.model(index=False, store=False, description="Model of User Favorites")
class UserFavorites(odm.Model):
    alert = odm.List(odm.Compound(Favorite), default=[], description="Alert page favorites")
    error = odm.List(odm.Compound(Favorite), default=[], description="Error page favorites")
    search = odm.List(odm.Compound(Favorite), default=[], description="Search page favorites")
    signature = odm.List(odm.Compound(Favorite), default=[], description="Signature page favorites")
    submission = odm.List(odm.Compound(Favorite), default=[], description="Submission page favorites")



import hashlib
import tlsh
from typing import Dict

from assemblyline.common import entropy

DEFAULT_BLOCKSIZE = 65536


# noinspection PyBroadException
def get_digests_for_file(path: str, blocksize: int = DEFAULT_BLOCKSIZE, calculate_entropy: bool = True,
                         on_first_block=lambda _b, _l, _p: {}) -> Dict:
    """ Generate digests for file reading only 'blocksize bytes at a time."""
    bc = None
    if calculate_entropy:
        try:
            bc = entropy.BufferedCalculator()
        except Exception:
            pass

    result = {}

    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()
    th = tlsh.Tlsh()
    size = 0

    with open(path, 'rb') as f:
        data = f.read(blocksize)
        length = len(data)

        if not size:
            result.update(on_first_block(data, length, path))

        while length > 0:
            if bc is not None:
                bc.update(data, length)
            md5.update(data)
            sha1.update(data)
            sha256.update(data)
            th.update(data)
            size += length

            data = f.read(blocksize)
            length = len(data)

    if bc is not None:
        result['entropy'] = bc.entropy()
    else:
        result['entropy'] = 0
    result['md5'] = md5.hexdigest()
    result['sha1'] = sha1.hexdigest()
    result['sha256'] = sha256.hexdigest()
    result['size'] = size

    # Try to finalise the TLSH Hash and add it to the results
    try:
        th.final()
        result['tlsh'] = th.hexdigest()
    except Exception:
        pass

    return result


def get_md5_for_file(path: str, blocksize: int = DEFAULT_BLOCKSIZE) -> str:
    md5 = hashlib.md5()
    with open(path, 'rb') as f:
        data = f.read(blocksize)
        length = len(data)

        while length > 0:
            md5.update(data)
            data = f.read(blocksize)
            length = len(data)

        return md5.hexdigest()


def get_sha256_for_file(path: str, blocksize: int = DEFAULT_BLOCKSIZE) -> str:
    sha256 = hashlib.sha256()
    with open(path, 'rb') as f:
        data = f.read(blocksize)
        length = len(data)

        while length > 0:
            sha256.update(data)
            data = f.read(blocksize)
            length = len(data)

        return sha256.hexdigest()



import hashlib
import json
import os
import random

from assemblyline.common import forge
from assemblyline.common.isotime import now_as_iso
from assemblyline.common.security import get_password_hash
from assemblyline.common.uid import get_random_id
from assemblyline.common.version import FRAMEWORK_VERSION, SYSTEM_VERSION, BUILD_MINOR
from assemblyline.odm.models.alert import Alert, Event, STATUSES, PRIORITIES
from assemblyline.odm.models.emptyresult import EmptyResult
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.file import File
from assemblyline.odm.models.heuristic import Heuristic
from assemblyline.odm.models.ontology import ResultOntology
from assemblyline.odm.models.result import Result
from assemblyline.odm.models.service import Service, UpdateSource
from assemblyline.odm.models.submission import Submission
from assemblyline.odm.models.user import TYPES, User
from assemblyline.odm.models.user_settings import UserSettings
from assemblyline.odm.models.safelist import Safelist
from assemblyline.odm.models.workflow import Workflow
from assemblyline.odm.randomizer import SERVICES, get_random_hash, random_minimal_obj, random_model_obj, \
    get_random_phrase, get_random_uri, get_random_word
from assemblyline.run.suricata_importer import SuricataImporter
from assemblyline.run.yara_importer import YaraImporter
from assemblyline.datastore.helper import AssemblylineDatastore

full_file_list = []
classification = forge.get_classification()


class NullLogger(object):
    def info(self, msg):
        pass

    def warn(self, msg):
        pass

    def error(self, msg):
        pass

    def exception(self, msg):
        pass

    def warning(self, msg):
        pass


def create_alerts(ds, alert_count=50, submission_list=None, log=None, workflows=[]):
    for _ in range(alert_count):
        a: Alert = random_model_obj(Alert)
        a.expiry_ts = now_as_iso(60 * 60 * 24 * 14)
        if isinstance(submission_list, list):
            submission = random.choice(submission_list)
            a.file.sha256 = submission.files[0].sha256
            a.sid = submission.sid

        a.owner = random.choice(['admin', 'user', 'other', None])
        if workflows:
            def generate_workflow_event(wf) -> Event:
                event: Event = random_minimal_obj(Event)
                if random.randint(0, 1) == 0:
                    # Overwrite with workflow information
                    event.entity_type = 'workflow'
                    event.entity_name = wf.name
                    event.entity_id = wf.workflow_id
                else:
                    # Overwrite with user information
                    event.entity_type = 'user'
                    event.entity_id = get_random_word()
                event.labels = [get_random_word() for _ in range(random.randint(0, 20))]
                event.status = random.choice(list(STATUSES) + [None])
                event.priority = random.choice(list(PRIORITIES) + [None])
                return event
            a.events = [generate_workflow_event(random.choice(workflows)) for _ in range(random.randint(0, 5))]

        # Clear sub-types
        for data_type in a.al.detailed.fields():
            if data_type in ['attrib']:
                continue
            for item in a.al.detailed[data_type]:
                item['subtype'] = None

        # Generate matching detailed IPs
        ips = []
        for ip in a.al.ip_static:
            ips.append(dict(subtype=None, type='network.static.ip', value=ip,
                            verdict=random.choice(['info', 'suspicious', 'malicious'])))
        for ip in a.al.ip_dynamic:
            ips.append(dict(subtype=None, type='network.dynamic.ip', value=ip,
                            verdict=random.choice(['info', 'suspicious', 'malicious'])))
        a.al.detailed.ip = ips
        a.al.ip = [ip['value'] for ip in ips]

        # Generate matching detailed Domains
        domains = []
        for dom in a.al.domain_static:
            domains.append(dict(subtype=None, type='network.static.domain', value=dom,
                                verdict=random.choice(['info', 'suspicious', 'malicious'])))
        for dom in a.al.domain_dynamic:
            domains.append(dict(subtype=None, type='network.dynamic.domain', value=dom,
                                verdict=random.choice(['info', 'suspicious', 'malicious'])))
        a.al.detailed.domain = domains
        a.al.domain = [domain['value'] for domain in domains]

        # Generate matching detailed URIs
        uris = []
        for uri in a.al.uri_static:
            uris.append(dict(subtype=None, type='network.static.uri', value=uri,
                             verdict=random.choice(['info', 'suspicious', 'malicious'])))
        for uri in a.al.uri_dynamic:
            uris.append(dict(subtype=None, type='network.dynamic.uri', value=uri,
                             verdict=random.choice(['info', 'suspicious', 'malicious'])))
        a.al.detailed.uri = uris
        a.al.uri = [uri['value'] for uri in uris]

        ds.alert.save(a.alert_id, a)
        if log:
            log.info(f"\t{a.alert_id}")

    ds.alert.commit()


def create_heuristics(ds, log=None):
    for srv in SERVICES.keys():
        for x in range(5):
            h = random_model_obj(Heuristic)
            h.heur_id = f"{srv.upper()}.{x + 1}"
            h.name = get_random_phrase()
            ds.heuristic.save(h.heur_id, h)
            if log:
                log.info(f'\t{h.heur_id}')

    ds.heuristic.commit()


def create_services(ds: AssemblylineDatastore, log=None, limit=None):
    if not limit:
        limit = len(SERVICES)

    for svc_name, svc in list(SERVICES.items())[:limit]:
        service_data = {
            "name": svc_name,
            "enabled": True,
            "category": svc[0],
            "stage": svc[1],
            "version": f"{FRAMEWORK_VERSION}.{SYSTEM_VERSION}.{BUILD_MINOR}.1",
            "docker_config": {
                "image": f"cccs/alsvc_{svc_name.lower()}:latest",
            },
        }

        if random.choice([True, False]):
            service_data['update_config'] = {
                "sources": [random_model_obj(UpdateSource)],
                "update_interval_seconds": 600,
                "generates_signatures": True
            }

        service_data = Service(service_data)
        for x in range(4):
            # Save the same service as v4
            service_data.version = f"{FRAMEWORK_VERSION}.{SYSTEM_VERSION}.{BUILD_MINOR}.{x+1}"
            ds.service.save(f"{service_data.name}_{service_data.version}", service_data)

        # Save the default delta entry
        ds.service_delta.save(service_data.name, {"version": service_data.version})
        if log:
            log.info(f'\t{svc_name}')

    ds.service_delta.commit()
    ds.service.commit()


def create_signatures(ds):
    yara = YaraImporter(logger=NullLogger())
    suricata = SuricataImporter(logger=NullLogger())
    signatures = yara.import_file(get_yara_sig_path(), source="YAR_SAMPLE", default_status="DEPLOYED")
    signatures.extend(suricata.import_file(get_suricata_sig_path(), source="ET_SAMPLE", default_status="DEPLOYED"))

    ds.signature.commit()

    return [s['name'] for s in signatures]


def _create_errors_for_file(ds, f, services_done, log=None):
    e_list = []
    for _ in range(random.randint(0, 1)):
        e = random_model_obj(Error)
        e.expiry_ts = now_as_iso(60 * 60 * 24 * 14)

        # Only one error per service per file
        while e.response.service_name in services_done:
            e.response.service_name = random.choice(list(SERVICES.keys()))
        services_done.append(e.response.service_name)

        # Set the sha256
        e.sha256 = f

        e_key = e.build_key()
        e_list.append(e_key)
        if log:
            log.info(f"\t\t\t{e_key}")
        ds.error.save(e_key, e)

    return e_list


def _create_results_for_file(ds, fs, f, possible_childs=None, log=None):
    r_list = []
    services_done = []
    section_body_format = ["TEXT", "MEMORY_DUMP", "GRAPH_DATA", "URL", "JSON", "KEY_VALUE"]
    section_depth_list = [[1, 1, 2, 3, 1], [1, 2, 1], [1, 2, 3, 1], [1, 2]]
    section_depth = random.choice(section_depth_list)
    for _ in range(random.randint(2, 5)):
        r = random_model_obj(Result)
        r.expiry_ts = now_as_iso(60 * 60 * 24 * 14)

        # Only one result per service per file
        while r.response.service_name in services_done:
            r.response.service_name = random.choice(list(SERVICES.keys()))

        for depth_id, section in enumerate(r.result.sections):
            section.depth = section_depth[depth_id % len(section_depth)]
            section.body_format = random.choice(section_body_format)
            section.heuristic.heur_id = random.choice([f"{r.response.service_name.upper()}.{x+1}" for x in range(5)])
            if section.body_format == "GRAPH_DATA":
                cmap_min = 0
                cmap_max = random.choice([5, 10, 20])
                color_map_data = {
                    'type': 'colormap',
                    'data': {
                        'domain': [cmap_min, cmap_max],
                        'values': [random.random() * cmap_max for _ in range(50)]
                    }
                }
                section.body = json.dumps(color_map_data)
            elif section.body_format == "URL":
                data = [{"url": get_random_uri()} for _ in range(random.randint(1, 4))]
                section.body = json.dumps(data)
            elif section.body_format in ["JSON", "KEY_VALUE"]:
                data = {get_random_word(): get_random_id() for _ in range(random.randint(3, 9))}
                section.body = json.dumps(data)

        services_done.append(r.response.service_name)

        # Set the sha256
        r.sha256 = f

        if random.randint(1, 10) > 8:
            # Generate and empty result
            r_key = f"{r.build_key()}.e"
            er = random_model_obj(EmptyResult)
            er.expiry_ts = now_as_iso(60 * 60 * 24 * 14)
            ds.emptyresult.save(r_key, er)
        else:
            r_key = r.build_key()
            # Set random extracted files that are not top level
            if not possible_childs:
                r.response.extracted = []
            else:
                for e in r.response.extracted:
                    e.sha256 = random.choice(possible_childs)

            # Set random supplementary files that are not top level
            if r.response.supplementary:
                # Edit the first file to be an ontology file
                s = r.response.supplementary[0]

                # Create a random ontology
                onto = random_minimal_obj(ResultOntology).as_primitives(strip_null=True)
                onto['file']['sha256'] = f
                onto['service'] = {
                    'name': r.response.service_name,
                    'version': r.response.service_version,
                    'tool_version': r.response.service_tool_version
                }

                # Create it's file record
                supp_file = random_model_obj(File)
                supp_file.expiry_ts = now_as_iso(60 * 60 * 24 * 14)
                byte_str = json.dumps(onto).encode('utf-8')
                sha256 = hashlib.sha256(byte_str).hexdigest()
                supp_file.sha256 = sha256
                ds.file.save(sha256, supp_file)
                fs.put(sha256, byte_str)

                # Add the random files
                s.sha256 = sha256
                s.name = "random.ontology"
                s.description = f"Random Ontology file for: {f}"

                r.response.supplementary = [s]

            ds.result.save(r_key, r)

        if log:
            log.info(f"\t\t\t{r_key}")
        r_list.append(r_key)

    return r_list


def create_submission(ds, fs, log=None):
    f_list = []
    r_list = []
    e_list = []

    first_level_files = []
    s = random_model_obj(Submission)
    s.expiry_ts = now_as_iso(60 * 60 * 24 * 14)

    if log:
        log.info(f"\t{s.sid}")
        log.info("\tGenerating files for submission...")
    for _ in range(random.randint(3, 6)):
        f = random_model_obj(File)
        f.expiry_ts = now_as_iso(60 * 60 * 24 * 14)
        byte_str = get_random_phrase(wmin=8, wmax=20).encode()
        sha256 = hashlib.sha256(byte_str).hexdigest()
        f.sha256 = sha256
        ds.file.save(sha256, f)
        fs.put(sha256, byte_str)

        if log:
            log.info(f"\t\t{sha256}")

        f_list.append(sha256)

    for _ in range(random.randint(1, 2)):
        first_level_files.append(f_list.pop())

    if log:
        log.info("\t\tGenerating results and errors for top level files...")
    for f in first_level_files:
        r_list.extend(_create_results_for_file(ds, fs, f, possible_childs=f_list, log=log))
        e_list.extend(_create_errors_for_file(ds, f, [x.split('.')[1] for x in r_list if x.startswith(f)], log=log))

    if log:
        log.info("\t\tGenerating results and errors for children files...")
    for f in f_list:
        r_list.extend(_create_results_for_file(ds, fs, f, log=log))
        e_list.extend(_create_errors_for_file(ds, f, [x.split('.')[1] for x in r_list if x.startswith(f)], log=log))

    s.results = r_list
    s.errors = e_list

    s.error_count = len(e_list)
    s.file_count = len({x[:64] for x in r_list})

    s.files = s.files[:len(first_level_files)]

    fid = 0
    for f in s.files:
        f.sha256 = first_level_files[fid]
        fid += 1

    s.params.psid = None
    s.state = 'completed'

    ds.submission.save(s.sid, s)

    ds.emptyresult.commit()
    ds.error.commit()
    ds.file.commit()
    ds.result.commit()
    ds.submission.commit()

    return s


def create_users(ds, log=None):
    admin_pass = os.getenv("DEV_ADMIN_PASS", 'admin') or 'admin'
    user_pass = os.getenv("DEV_USER_PASS", 'user') or 'user'
    user_data = User({
        "agrees_with_tos": "NOW",
        "apikeys": {'devkey': {'acl': ["R", "W"], "password": get_password_hash(admin_pass)}},
        "classification": classification.RESTRICTED,
        "name": "Administrator",
        "email": "admin@assemblyline.cyber.gc.ca",
        "password": get_password_hash(admin_pass),
        "uname": "admin",
        "type": [TYPES.admin]})
    ds.user.save('admin', user_data)
    ds.user_settings.save('admin', UserSettings({"ignore_cache": True, "deep_scan": True}))
    if log:
        log.info(f"\tU:{user_data.uname}   P:{admin_pass}")

    user_data = User({
        "name": "User",
        "email": "user@assemblyline.cyber.gc.ca",
        "apikeys": {'devkey': {'acl': ["R", "W"], "password": get_password_hash(user_pass)}},
        "password": get_password_hash(user_pass),
        "uname": "user",
        "type": [TYPES.user]})
    ds.user.save('user', user_data)
    ds.user_settings.save('user', UserSettings())
    if log:
        log.info(f"\tU:{user_data.uname}   P:{user_pass}")

    ds.user.commit()


def create_safelists(ds, log=None):
    for _ in range(20):
        sl = random_model_obj(Safelist, as_json=True)
        if sl['type'] == 'file':
            sl.pop('tag', None)
        elif sl['type'] == 'tag':
            sl.pop('file', None)
        sl['hashes']['sha256'] = "0" + get_random_hash(63)
        ds.safelist.save(sl['hashes']['sha256'], sl)
        if log:
            log.info(f"\t{sl['hashes']['sha256']}")

    ds.safelist.commit()


def create_workflows(ds, log=None):
    workflows = []
    for _ in range(20):
        w_id = get_random_id()
        workflow = random_model_obj(Workflow)
        workflow.workflow_id = w_id
        ds.workflow.save(w_id, workflow)
        if log:
            log.info(f'\t{w_id}')
        workflows.append(workflow)

    ds.workflow.commit()
    return workflows


def get_suricata_sig_path():
    for (d, _, filenames) in os.walk(__file__[:-11]):
        for f in filenames:
            if f == 'sample_suricata.rules':
                return os.path.join(d, f)

    raise Exception('Could not find suricata sample file...')


def get_yara_sig_path():
    for (d, _, filenames) in os.walk(__file__[:-11]):
        for f in filenames:
            if f == 'sample_rules.yar':
                return os.path.join(d, f)

    raise Exception('Could not find yara sample file...')


def wipe_alerts(ds):
    ds.alert.wipe()


def wipe_heuristics(ds):
    ds.heuristic.wipe()


def wipe_services(ds):
    ds.service.wipe()
    ds.service_delta.wipe()


def wipe_signatures(ds):
    ds.signature.wipe()


def wipe_submissions(ds, fs):
    ds.error.wipe()
    ds.file.wipe()
    ds.emptyresult.wipe()
    ds.result.wipe()
    ds.submission.wipe()
    ds.submission_summary.wipe()
    ds.submission_tree.wipe()

    for f in full_file_list:
        fs.delete(f)


def wipe_users(ds):
    ds.user.wipe()
    ds.user_settings.wipe()
    ds.user_avatar.wipe()
    ds.user_favorites.wipe()


def wipe_safelist(ds):
    ds.safelist.wipe()


def wipe_workflows(ds):
    ds.workflow.wipe()

def wipe_all_except_users(ds, fs):
    wipe_alerts(ds)
    wipe_heuristics(ds)
    wipe_services(ds)
    wipe_signatures(ds)
    wipe_submissions(ds, fs)
    wipe_safelist(ds)
    wipe_workflows(ds)



import logging
import re
import struct
import subprocess
import sys
import threading
import uuid
import zipfile
from binascii import hexlify
from typing import Dict, Optional, Tuple, Union

import magic
import msoffcrypto
import ssdeep
import yaml
import yara
from assemblyline.common.digests import get_digests_for_file
from assemblyline.common.forge import get_cachestore, get_config, get_constants, get_datastore
from assemblyline.common.identify_defaults import OLE_CLSID_GUIDs
from assemblyline.common.identify_defaults import magic_patterns as default_magic_patterns
from assemblyline.common.identify_defaults import trusted_mimes as default_trusted_mimes
from assemblyline.common.str_utils import dotdump, safe_str
from assemblyline.filestore import FileStoreException
from assemblyline.remote.datatypes.events import EventWatcher
from cart import get_metadata_only

constants = get_constants()


# These headers are found in the custom.magic file to assist with identification, and are imported by services
# that can create files with a high-confidence type
CUSTOM_PS1_ID = b"#!/usr/bin/env pwsh\n"
CUSTOM_BATCH_ID = b"REM Batch extracted by Assemblyline\n"


class Identify():
    def __init__(self, use_cache: bool = True, config=None, datastore=None, log=None) -> None:
        self.log = log or logging.getLogger('assemblyline.identify')
        self.config = None
        self.datastore = None
        self.use_cache = use_cache
        self.custom = re.compile(r"^custom: ", re.IGNORECASE)
        self.lock = threading.Lock()
        self.yara_default_externals = {'mime': '', 'magic': '', 'type': ''}

        # If cache is use, load the config and datastore objects to load potential items from cache
        if self.use_cache:
            self.log.info("Using cache with identify")
            self.config = config or get_config()
            self.datastore = datastore or get_datastore(config)

        # Load all data for the first time
        self._load_magic_file()
        self._load_yara_file()
        self._load_magic_patterns()
        self._load_trusted_mimes()

        # Register hot reloader
        if self.use_cache:
            self.reload_map = {
                'magic': self._load_magic_file,
                'mimes': self._load_trusted_mimes,
                'patterns': self._load_magic_patterns,
                'yara': self._load_yara_file
            }
            self.reload_watcher: Optional[EventWatcher[str]] = EventWatcher()
            self.reload_watcher.register('system.identify', self._handle_reload_event)
            self.reload_watcher.start()
        else:
            self.reload_watcher = None
            self.reload_map = {}

    def _handle_reload_event(self, data: Optional[str]):
        if data is None:
            # handle disconnect event, we may be out of sync
            for reload_func in self.reload_map.values():
                reload_func()
        else:
            # update information
            func = self.reload_map.get(data, None)
            if func is not None:
                func()
            else:
                self.log.error(f"Invalid system.identify message received: {data}")

    def _load_magic_patterns(self):
        self.magic_patterns = default_magic_patterns

        if self.use_cache:
            self.log.info("Checking for custom magic patterns...")
            with get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
                try:
                    patterns = cache.get('custom_patterns')
                    if patterns:
                        self.magic_patterns = yaml.safe_load(patterns)
                        self.log.info("Custom magic patterns loaded!")
                    else:
                        self.log.info("No custom magic patterns found.")
                except FileStoreException:
                    self.log.info("No custom magic patterns found.")

        compiled_patterns = [[x['al_type'], re.compile(x['regex'], re.IGNORECASE)] for x in self.magic_patterns]
        with self.lock:
            self.compiled_magic_patterns = compiled_patterns

    def _load_trusted_mimes(self):
        trusted_mimes = default_trusted_mimes

        if self.use_cache:
            self.log.info("Checking for custom trusted mimes...")
            with get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
                try:
                    mimes = cache.get('custom_mimes')
                    if mimes:
                        trusted_mimes = yaml.safe_load(mimes)
                        self.log.info("Custom trusted mimes loaded!")
                    else:
                        self.log.info("No custom magic patterns found.")
                except FileStoreException:
                    self.log.info("No custom trusted mimes found.")

        with self.lock:
            self.trusted_mimes = trusted_mimes

    def _load_magic_file(self):
        self.magic_file = ':'.join((constants.MAGIC_RULE_PATH, '/usr/share/file/magic.mgc'))

        if self.use_cache:
            self.log.info("Checking for custom magic file...")
            with get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
                try:
                    custom_magic = "/tmp/custom.magic"
                    cache.download('custom_magic', custom_magic)
                    self.magic_file = ':'.join((custom_magic, '/usr/share/file/magic.mgc'))
                    self.log.info("Custom magic file loaded!")
                except FileStoreException:
                    self.log.info("No custom magic file found.")

        with self.lock:
            self.file_type = magic.magic_open(magic.MAGIC_CONTINUE + magic.MAGIC_RAW)
            magic.magic_load(self.file_type, self.magic_file)

            self.mime_type = magic.magic_open(magic.MAGIC_CONTINUE + magic.MAGIC_RAW + magic.MAGIC_MIME)
            magic.magic_load(self.mime_type, self.magic_file)

    def _load_yara_file(self):
        self.yara_file = constants.YARA_RULE_PATH

        if self.use_cache:
            self.log.info("Checking for custom yara file...")
            with get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
                try:
                    custom_yara = "/tmp/custom.yara"
                    cache.download('custom_yara', custom_yara)
                    self.yara_file = custom_yara
                    self.log.info("Custom yara file loaded!")
                except FileStoreException:
                    self.log.info("No custom magic file found.")

        yara_rules = yara.compile(filepaths={"default": self.yara_file}, externals=self.yara_default_externals)
        with self.lock:
            self.yara_rules = yara_rules

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    def stop(self):
        if self.reload_watcher:
            self.reload_watcher.stop()

    def ident(self, buf, length: int, path) -> Dict:
        data = {"ascii": None, "hex": None, "magic": None, "mime": None, "type": "unknown"}

        if length <= 0:
            return data

        header = buf[: min(64, length)]
        data["ascii"] = dotdump(header)
        data["hex"] = safe_str(hexlify(header))

        # noinspection PyBroadException
        try:
            # Loop over the labels returned by libmagic, ...
            labels = []
            mimes = []

            with self.lock:
                try:
                    labels = magic.magic_file(self.file_type, path).split(b"\n")
                except magic.MagicException as me:
                    labels = me.message.split(b"\n")

                try:
                    mimes = magic.magic_file(self.mime_type, path).split(b"\n")
                except magic.MagicException as me:
                    mimes = me.message.split(b"\n")

            mimes = [mime[2:].strip() if mime.startswith(b"- ") else mime.strip() for mime in mimes]
            labels = [label[2:].strip() if label.startswith(b"- ") else label.strip() for label in labels]

            # For user feedback set the mime and magic meta data to always be the primary
            # libmagic responses
            if len(labels) > 0:

                def find_special_words(word, labels):
                    for index, label in enumerate(labels):
                        if word in label:
                            return index
                    return -1

                # If an expected label is not the first label returned by Magic, then make it so
                # Manipulating the mime accordingly varies between special word cases
                special_word_cases = [
                    (b"OLE 2 Compound Document : Microsoft Word Document", False),
                    (b"Lotus 1-2-3 WorKsheet", True),
                ]
                for word, alter_mime in special_word_cases:
                    index = find_special_words(word, labels)
                    if index >= 0:
                        labels.insert(0, labels.pop(index))
                        if len(labels) == len(mimes) and alter_mime:
                            mimes.insert(0, mimes.pop(index))
                data["magic"] = safe_str(labels[0])

            for mime in mimes:
                if mime != b"":
                    data["mime"] = safe_str(mime)
                    break

            # First lets try to find any custom types
            for label in labels:
                label = dotdump(label)

                if self.custom.match(label):
                    # Some things, like executable have additional data appended to their identification, like
                    # ", dynamically linked, stripped" that we do not want to use as part of the type.
                    data["type"] = label.split("custom: ")[1].split(",", 1)[0].strip()
                    break

            # Second priority is mime times marked as trusted
            if data["type"] == "unknown":
                with self.lock:
                    trusted_mimes = self.trusted_mimes

                for mime in mimes:
                    mime = dotdump(mime)

                    if mime in trusted_mimes:
                        data["type"] = trusted_mimes[mime]
                        break

            # As a third priority try matching the magic_patterns
            if data["type"] == "unknown":
                found = False
                with self.lock:
                    compiled_magic_patterns = self.compiled_magic_patterns

                for label in labels:
                    for entry in compiled_magic_patterns:
                        if entry[1].search(dotdump(label)):  # pylint: disable=E1101
                            data['type'] = entry[0]
                            found = True
                            break
                    if found:
                        break

        except Exception as e:
            self.log.error(f"An error occured during file identification: {e.__class__.__name__}({str(e)})")
            pass

        # If mime is text/* and type is unknown, set text/plain to trigger
        # language detection later.
        if data["type"] == "unknown" and data['mime'] is not None and data['mime'].startswith("text/"):
            data["type"] = "text/plain"

        # Lookup office documents by GUID if we're still not sure what they are
        if data["type"] == "document/office/unknown":
            # noinspection PyBroadException
            try:
                root_entry_property_offset = buf.find(u"Root Entry".encode("utf-16-le"))
                if -1 != root_entry_property_offset:
                    # Get root entry's GUID and try to guess document type
                    clsid_offset = root_entry_property_offset + 0x50
                    if len(buf) >= clsid_offset + 16:
                        clsid = buf[clsid_offset: clsid_offset + 16]
                        if len(clsid) == 16 and clsid != b"\0" * len(clsid):
                            clsid_str = uuid.UUID(bytes_le=clsid)
                            clsid_str = clsid_str.urn.rsplit(":", 1)[-1].upper()
                            if clsid_str in OLE_CLSID_GUIDs:
                                data["type"] = OLE_CLSID_GUIDs[clsid_str]
                        else:
                            bup_details_offset = buf[
                                : root_entry_property_offset + 0x100
                            ].find(u"Details".encode("utf-16-le"))
                            if -1 != bup_details_offset:
                                data["type"] = "quarantine/mcafee"
            except Exception:
                pass

        return data

    def yara_ident(self, path: str, info: Dict, fallback="unknown") -> Tuple[str, Union[str, int]]:
        externals = {k: v or "" for k, v in info.items() if k in self.yara_default_externals}
        try:
            with self.lock:
                yara_rules = self.yara_rules
            matches = yara_rules.match(path, externals=externals, fast=True)
            matches.sort(key=lambda x: x.meta.get('score', 0), reverse=True)
            for match in matches:
                ftype = match.meta.get("type", None)
                if ftype:
                    return ftype
        except Exception as e:
            self.log.warning(f"Yara file identifier failed with error: {str(e)}")
            matches = []

        return fallback

    def fileinfo(self, path: str) -> Dict:
        path = safe_str(path)
        data = get_digests_for_file(path, on_first_block=self.ident)
        data["ssdeep"] = ssdeep.hash_from_file(path)

        # Check if file empty
        if not int(data.get("size", -1)):
            data["type"] = "empty"

        # Futher identify zip files based of their content
        elif data["type"] in ["archive/zip", "java/jar", "document/office/unknown"]:
            data["type"] = zip_ident(path, data["type"])

        # Further check CaRT files, they may have an explicit type set
        elif data["type"] == "archive/cart":
            data["type"] = cart_ident(path)

        # Further identify dos executables has this may be a PE that has been misidentified
        elif data["type"] == "executable/windows/dos":
            data["type"] = dos_ident(path)

        # If we're so far failed to identified the file, lets run the yara rules
        elif "unknown" in data["type"] or data["type"] == "text/plain":
            data["type"] = self.yara_ident(path, data, fallback=data["type"])

        # Extra checks for office documents
        #  - Check for encryption
        if data["type"] in [
            "document/office/word",
            "document/office/excel",
            "document/office/powerpoint",
            "document/office/unknown",
        ]:
            try:
                msoffcrypto_obj = msoffcrypto.OfficeFile(open(path, "rb"))
                if msoffcrypto_obj and msoffcrypto_obj.is_encrypted():
                    data["type"] = "document/office/passwordprotected"
            except Exception:
                # If msoffcrypto can't handle the file to confirm that it is/isn't password protected,
                # then it's not meant to be. Moving on!
                pass

        # Extra checks for PDF documents
        #  - Check for encryption
        #  - Check for PDF collection (portfolio)
        if data["type"] == "document/pdf":
            # Password protected documents typically contain '/Encrypt'
            pdf_content = open(path, "rb").read()
            if re.search(b"/Encrypt", pdf_content):
                data["type"] = "document/pdf/passwordprotected"
            # Portfolios typically contain '/Type/Catalog/Collection
            elif re.search(b"/Type/Catalog/Collection", pdf_content):
                data["type"] = "document/pdf/portfolio"

        return data


def zip_ident(path: str, fallback: str) -> str:
    file_list = []

    try:
        with zipfile.ZipFile(path, "r") as zf:
            file_list = [zfname for zfname in zf.namelist()]
    except Exception:
        try:
            stdout, _ = subprocess.Popen(
                ["unzip", "-l", path], stderr=subprocess.PIPE, stdout=subprocess.PIPE
            ).communicate()
            lines = stdout.splitlines()
            index = lines[1].index(b"Name")
            for file_name in lines[3:-2]:
                file_list.append(safe_str(file_name[index:]))
        except Exception:
            return fallback

    tot_files = 0
    tot_class = 0
    tot_jar = 0

    is_ipa = False
    is_jar = False
    is_word = False
    is_excel = False
    is_ppt = False
    doc_props = False
    doc_rels = False
    doc_types = False
    android_manifest = False
    android_dex = False

    for file_name in file_list:
        if file_name.startswith("META-INF/"):
            is_jar = True
        elif file_name == "AndroidManifest.xml":
            android_manifest = True
        elif file_name == "classes.dex":
            android_dex = True
        elif file_name.startswith("Payload/") and file_name.endswith(".app/Info.plist"):
            is_ipa = True
        elif file_name.endswith(".class"):
            tot_class += 1
        elif file_name.endswith(".jar"):
            tot_jar += 1
        elif file_name.startswith("word/"):
            is_word = True
        elif file_name.startswith("xl/"):
            is_excel = True
        elif file_name.startswith("ppt/"):
            is_ppt = True
        elif file_name.startswith("docProps/"):
            doc_props = True
        elif file_name.startswith("_rels/"):
            doc_rels = True
        elif file_name == "[Content_Types].xml":
            doc_types = True

        tot_files += 1

    if 0 < tot_files < (tot_class + tot_jar) * 2:
        is_jar = True

    if is_jar and android_manifest and android_dex:
        return "android/apk"
    elif is_ipa:
        return "ios/ipa"
    elif is_jar:
        return "java/jar"
    elif (doc_props or doc_rels) and doc_types:
        if is_word:
            return "document/office/word"
        elif is_excel:
            return "document/office/excel"
        elif is_ppt:
            return "document/office/powerpoint"
        else:
            return "document/office/unknown"
    else:
        return "archive/zip"


# noinspection PyBroadException
def cart_ident(path: str) -> str:
    try:
        metadata = get_metadata_only(path)
    except Exception:
        return "corrupted/cart"
    return metadata.get("al", {}).get("type", "archive/cart")


def dos_ident(path: str) -> str:
    # noinspection PyBroadException
    try:
        with open(path, "rb") as fh:
            file_header = fh.read(0x40)
            if file_header[0:2] != b"MZ":
                raise ValueError()

            (header_pos,) = struct.unpack("<I", file_header[-4:])
            fh.seek(header_pos)
            if fh.read(4) != b"PE\x00\x00":
                raise ValueError()
            (machine_id,) = struct.unpack("<H", fh.read(2))
            if machine_id == 0x014C:
                width = 32
            elif machine_id == 0x8664:
                width = 64
            else:
                raise ValueError()
            (characteristics,) = struct.unpack("<H", fh.read(18)[-2:])
            if characteristics & 0x2000:
                pe_type = "dll"
            elif characteristics & 0x0002:
                pe_type = "pe"
            else:
                raise ValueError()
            return "executable/windows/%s%i" % (pe_type, width)
    except Exception:
        pass
    return "executable/windows/dos"


if __name__ == "__main__":
    from pprint import pprint

    use_cache = True
    args = sys.argv[1:]
    if "--no-cache" in args:
        args.remove("--no-cache")
        use_cache = False

    identify = Identify(use_cache=use_cache)

    if len(args) > 0:
        pprint(identify.fileinfo(args[0]))
    else:
        name = sys.stdin.readline().strip()
        while name:
            a = identify.fileinfo(name)
            print(
                "\t".join(
                    dotdump(str(a[k]))
                    for k in (
                        "type",
                        "ascii",
                        "entropy",
                        "hex",
                        "magic",
                        "mime",
                        "md5",
                        "sha1",
                        "sha256",
                        "ssdeep",
                        "size",
                    )
                )
            )
            name = sys.stdin.readline().strip()



from __future__ import annotations
from typing import Any, Callable, Optional, TYPE_CHECKING, TypeVar, Generic
import json
import logging
import threading

from assemblyline.remote.datatypes import retry_call, get_client

if TYPE_CHECKING:
    from redis import Redis
    from redis.client import PubSub


logger = logging.getLogger(__name__)

MessageType = TypeVar('MessageType')


class EventSender(Generic[MessageType]):
    def __init__(self, prefix: str, host=None, port=None, private=None,
                 serializer: Callable[[MessageType],
                                      str] = json.dumps):
        self.client: Redis[Any] = get_client(host, port, private)
        self.prefix = prefix.lower()
        if not self.prefix.endswith('.'):
            self.prefix += '.'
        self.serializer = serializer

    def send(self, name: str, data: MessageType):
        path = self.prefix + name.lower().lstrip('.')
        retry_call(self.client.publish, path, self.serializer(data))


class PubSubWorkerThread(threading.Thread):
    """
    Worker thread that continually reads messages from pubsub.

    We reimplement the worker thread here rather than use the one in the redis
    package because we want to use the subscribe messages for disconnect/reconnect
    detection.
    """

    def __init__(self, watcher, exception_handler=None, skip_first_refresh=True):
        super().__init__(daemon=True)
        self.watcher = watcher
        self.exception_handler = exception_handler
        self._running = threading.Event()
        self.skip_first_refresh = skip_first_refresh

    def run(self):
        if self._running.is_set():
            return
        self._running.set()

        pubsub: PubSub = self.watcher.pubsub
        ping = self.watcher.client.ping
        sleep_time = 1
        initialized = set()

        while self._running.is_set():
            try:
                message = pubsub.get_message(ignore_subscribe_messages=False, timeout=sleep_time)
                if message is not None and message['type'] == 'psubscribe':
                    channel = message.get("channel")
                    if channel is None:
                        continue

                    if self.skip_first_refresh and channel not in initialized:
                        initialized.add(channel)
                        continue

                    handler = pubsub.patterns.get(channel, None)
                    if handler:
                        handler(None)

            except BaseException as exception:
                # Present the error
                logger.error(f"Exception in pubsub watcher: {exception}")

                # Wait until we can reach the server
                retry_call(ping)

        pubsub.close()

    def stop(self):
        # trip the flag so the run loop exits. the run loop will
        # close the pubsub connection, which disconnects the socket
        # and returns the connection to the pool.
        self._running.clear()


class EventWatcher(Generic[MessageType]):
    def __init__(self, host=None, port=None, deserializer: Callable[[str], MessageType] = json.loads):
        self.client: Redis[Any] = get_client(host, port, False)
        self.pubsub = retry_call(self.client.pubsub)
        self.pubsub.ignore_subscribe_messages = False
        self.thread: Optional[PubSubWorkerThread] = None
        self.deserializer = deserializer
        self.skip_first_refresh = True

    def register(self, path: str, callback: Callable[[MessageType | None], None]):
        def _callback(message: Optional[dict[str, Any]]):
            if message is None:
                callback(None)
            elif message['type'] == 'pmessage':
                data = self.deserializer(message.get('data', ''))
                callback(data)
        self.pubsub.psubscribe(**{path.lower(): _callback})

    def start(self):
        self.thread = PubSubWorkerThread(self, skip_first_refresh=self.skip_first_refresh)
        self.thread.start()
        return self.thread

    def stop(self):
        if self.thread is not None:
            self.thread.stop()



#!/usr/bin/env python

import sys

from assemblyline.remote.datatypes.queues.comms import CommsQueue
from pprint import pprint


if __name__ == "__main__":
    queue_name = None
    if len(sys.argv) > 1:
        queue_name = sys.argv[1]

    if queue_name is None:
        print("\nERROR: You must specify a queue name.\n\npubsub_reader.py [queue_name]")
        exit(1)

    print(f"Listening for messages on '{queue_name}' queue.")

    q = CommsQueue(queue_name)

    try:
        while True:
            for msg in q.listen():
                pprint(msg)
    except KeyboardInterrupt:
        print('Exiting')
    finally:
        q.close()



from __future__ import annotations
import hashlib
from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()

SUBMISSION_STATES = ['failed', 'submitted', 'completed']
DEFAULT_SRV_SEL = ["Filtering", "Antivirus", "Static Analysis", "Extraction", "Networking"]
DEFAULT_RESUBMIT = []


@odm.model(index=True, store=False, description="File Model of Submission")
class File(odm.Model):
    name = odm.Keyword(copyto="__text__", description="Name of the file")
    size = odm.Optional(odm.Integer(), description="Size of the file in bytes")
    sha256 = odm.SHA256(copyto="__text__", description="SHA256 hash of the file")


@odm.model(index=False, store=False, description="Service Selection Scheme")
class ServiceSelection(odm.Model):
    selected = odm.List(odm.Keyword(), default=DEFAULT_SRV_SEL, description="List of selected services")
    excluded = odm.List(odm.Keyword(), default=[], description="List of excluded services")
    rescan = odm.List(
        odm.Keyword(),
        default=[],
        description="List of services to rescan when initial run scores as malicious")
    resubmit = odm.List(odm.Keyword(), default=DEFAULT_RESUBMIT,
                        description="Add to service selection when resubmitting")
    runtime_excluded = odm.List(odm.Keyword(), default=[], description="List of runtime excluded services")


# Fields in the parameters used to calculate hashes used for result caching
_KEY_HASHED_FIELDS = {
    'classification',
    'deep_scan',
    'ignore_cache',
    'ignore_dynamic_recursion_prevention',
    'ignore_filtering',
    'ignore_size',
    'max_extracted',
    'max_supplementary',
}


@odm.model(index=True, store=False, description="Submission Parameters")
class SubmissionParams(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED,
                                        description="Original classification of the submission")
    deep_scan = odm.Boolean(default=False, description="Should a deep scan be performed?")
    description = odm.Text(store=True, copyto="__text__", description="Description of the submission")
    generate_alert = odm.Boolean(default=False, description="Should this submission generate an alert?")
    groups = odm.List(odm.Keyword(), default=[], description="List of groups related to this scan")
    ignore_cache = odm.Boolean(default=False, description="Ignore the cached service results?")
    ignore_dynamic_recursion_prevention = odm.Boolean(
        default=False, description="Should we ignore dynamic recursion prevention?")
    ignore_filtering = odm.Boolean(default=False, description="Should we ignore filtering services?")
    ignore_size = odm.Boolean(default=False, description="Ignore the file size limits?")
    never_drop = odm.Boolean(default=False, description="Exempt from being dropped by ingester?")
    malicious = odm.Boolean(default=False, description="Is the file submitted already known to be malicious?")
    max_extracted = odm.Integer(default=500, description="Max number of extracted files")
    max_supplementary = odm.Integer(default=500, description="Max number of supplementary files")
    priority = odm.Integer(default=1000, description="Priority of the scan")
    profile = odm.Boolean(default=False, description="Should the submission do extra profiling?")
    psid = odm.Optional(odm.UUID(), description="Parent submission ID")
    quota_item = odm.Boolean(default=False, description="Does this submission count against quota?")
    services = odm.Compound(ServiceSelection, default={}, description="Service selection")
    service_spec = odm.Mapping(odm.Mapping(odm.Any()), default={}, index=False, store=False,
                               description="Service-specific parameters")
    submitter = odm.Keyword(store=True, copyto="__text__", description="User who submitted the file")
    ttl = odm.Integer(default=0, description="Time, in days, to live for this submission")
    type = odm.Keyword(default="USER", description="Type of submission")
    initial_data = odm.Optional(odm.Text(index=False), description="Initialization for temporary submission data")
    auto_archive = odm.Boolean(default=False,
                               description="Does the submission automatically goes into the archive when completed?")
    delete_after_archive = odm.Boolean(
        default=False,
        description="When the submission is archived, should we delete it from hot storage right away?")

    def get_hashing_keys(self):
        """Get the sections of the submission parameters that should be used in result hashes."""
        data = self.as_primitives()
        return {k: v for k, v in data.items() if k in _KEY_HASHED_FIELDS}

    def create_filescore_key(self, sha256, services: list = None):
        """This is the key used to store the final score of a submission for fast lookup.

        This lookup is one of the methods used to check for duplication in ingestion process,
        so this key is fairly sensitive.
        """
        # TODO do we need this version thing still be here?
        # One up this if the cache is ever messed up and we
        # need to quickly invalidate all old cache entries.
        version = 0

        if services is None:
            services = self.services.selected

        data = self.get_hashing_keys()
        data['service_spec'] = sorted((key, sorted(values.items())) for key, values in self.service_spec.items())
        data['sha256'] = sha256
        data['services'] = [str(x) for x in services]

        s = ', '.join([f"{k}: {data[k]}" for k in sorted(data.keys())])

        return 'v'.join([str(hashlib.md5(s.encode()).hexdigest()), str(version)])


@odm.model(index=True, store=True, description="Submission-Relevant Times")
class Times(odm.Model):
    completed = odm.Optional(odm.Date(store=False), description="Date at which the submission finished scanning")
    submitted = odm.Date(default="NOW", description="Date at which the submission started scanning")


@odm.model(index=True, store=False, description="Submission Verdict")
class Verdict(odm.Model):
    malicious = odm.List(odm.Keyword(), default=[], description="List of user that thinks this submission is malicious")
    non_malicious = odm.List(
        odm.Keyword(),
        default=[],
        description="List of user that thinks this submission is non-malicious")


@odm.model(index=True, store=True, description="Model of Submission")
class Submission(odm.Model):
    archive_ts = odm.Optional(odm.Date(store=False, description="Archiving timestamp (Deprecated)"))
    archived = odm.Boolean(default=False, description="Document is present in the malware archive")
    classification = odm.Classification(description="Classification of the submission")
    error_count = odm.Integer(description="Total number of errors in the submission")
    errors: list[str] = odm.List(odm.Keyword(), store=False, description="List of error keys")
    expiry_ts = odm.Optional(odm.Date(store=False), description="Expiry timestamp")
    file_count = odm.Integer(description="Total number of files in the submission")
    files: list[File] = odm.List(odm.Compound(File), description="List of files that were originally submitted")
    max_score = odm.Integer(description="Maximum score of all the files in the scan")
    metadata = odm.FlattenedObject(store=False, description="Metadata associated to the submission")
    params: SubmissionParams = odm.Compound(SubmissionParams, description="Submission parameter details")
    results: list[str] = odm.List(odm.Keyword(), store=False, description="List of result keys")
    sid: str = odm.UUID(copyto="__text__", description="Submission ID")
    state = odm.Enum(values=SUBMISSION_STATES, description="Status of the submission")
    to_be_deleted = odm.Boolean(
        default=False, description="This document is going to be deleted as soon as it finishes")
    times = odm.Compound(Times, default={}, description="Submission-specific times")
    verdict = odm.Compound(Verdict, default={}, description="Malicious verdict details")
    from_archive = odm.Boolean(index=False, default=False, description="Was loaded from the archive")

    # the filescore key, used in deduplication. This is a non-unique key, that is
    # shared by submissions that may be processed as duplicates.
    scan_key = odm.Optional(odm.Keyword(store=False, index=False))

    def is_submit(self):
        return self.state == 'submitted'

    def is_complete(self):
        return self.state == 'completed'

    def is_initial(self):
        return self.is_submit() and not self.params.psid



from assemblyline import odm
from assemblyline.common import forge
Classification = forge.get_classification()


@odm.model(index=True, store=False, description="Submission Tree Model")
class SubmissionTree(odm.Model):
    classification = odm.Classification(default=Classification.UNRESTRICTED, description="Classification of the cache")
    filtered = odm.Boolean(default=False, description="Has this cache entry been filtered?")
    expiry_ts = odm.Date(description="Expiry timestamp")
    supplementary = odm.Text(index=False, description="Tree of supplementary files")
    tree = odm.Text(index=False, description="File tree cache")
